/**
 *	PhilEver Read/Hi-level draw.  
 *	Resurrected by Doug Nachand 7/6/01.
 *	@file topo.h
 *	@author Glenn Chappell
 *	@date 5/18/93
 */
#ifndef TOPO_H
#define TOPO_H

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "level.h"

extern void calcsquare(int borders, float rows[16][4], struct patchdata **data);
extern void calccap();
extern void calcsaddle(float thetrs[32][4], struct patchdata **data);
extern void drawsquare(struct patchdata *data);
//extern void drawcap(struct patchdata *data);
extern void drawsaddle(struct patchdata *data);

struct patchrec
{
	int patchtype;		///< 0 = square, 1 = cap, 2 = saddle
	int patchflags;
	int numpatchpts;
	int *levs;
	int *tows;
	struct patchdata *pd;
};

struct patchlist
{
	int patchnum;
	struct patchlist *next;
};

struct vert
{
	int left;	///< Left neighbor of this vertex
	int right;	///< Right neighbor of this vertex
	int up_or;
	int down_or;
	int point;	///< Holds the place(i.e. first, second, etc) where the point is in the sps file
};

/// This holds the connectedness of each control point.  
//The first dimension is the level and the second is the number of the control point from the sps file.
struct vert verts[100][100];		
// This holds the first number in the cycle.  The first dimension
// is the level and the second is the cycle in that level.
int cyclestarts[100][100];			
// This holds the size of the cycle from the second dimension 
// of the array in the level denoted by the first dimension.
int cyclesizes[100][100];			
int numcycles[100];	///< Holds the number of cycles for each level
int numlevels;		///< Total number of levels
/// Holds the control points.  The first dimension is the level, the second 
/// is the point on the level, and the third are the four coords, x, y, z, and time?
float points[100][100][4];			
struct patchlist *pls[100][100];
int ptsperlevel[100];	///< Holds the total number of points for each level
struct patchrec *patches[1000];
int numpatches;		///< The total number of patches in our model
int maxtowernum;

/**
 *  Function gets the pointer to the file we want.
 *	@param filename The filename we want to open
 *	@param write Flag for writing permissions on the file (0=read-only, 1=read and write)
 *  @return fileptr A pointer to an input file
 *	FILE structure type : Stores information about current state 
 *	  of stream; used in all stream I/O operations.  Declared in stdio.h
 */
FILE *ropen(char *filename, int write)
{
	FILE *fileptr;
	
	if (!strcmp(filename, "-"))	// if the filename and - are the same, get input from the console?
		return(stdin);
	if(!write)
	{
		if ((fileptr = fopen(filename, "r")) == NULL) 
		{
			fprintf(stderr, "%s: Cannot read file \"%s\"\n", "myname?", filename); /* !! */
			exit(1);
		}
	}
	else
	{
		if((fileptr = fopen(filename, "w")) == NULL) // open for reading and writing
		{
			fprintf(stderr, "%s: Cannot read file \"%s\"\n","myname?", filename); /* !! */
			exit(1);
		}
	}
	return(fileptr);
}

/**
 *	Closes the file stream
 *	@param fileptr Pointer to the file stream.
 */
void rclose(FILE *fileptr)
{
	if (fileptr != stdin)
		fclose(fileptr);
}

/**
 *	Allocate space
 *	@param size Size of the space we want to allocate
 *	@return ptr A pointer to the space we just allocated
 */
void* allocate(long size)
{
	void *ptr;
	
	if ((ptr = (void*)malloc(size)) == NULL) 
	{
		fprintf(stderr, "%s: Out of memory\n", "myname?"); /* !! */
		exit(1);
	}
	return(ptr);
}


FILE *tokenfile;	///< pointer to the input we are using
int firsttoken;		///< flag to say we haven't looked at a token yet
int tokenerr;
char tok[100];

void tokensetup(FILE *fp)
{
	tokenfile = fp;
	firsttoken = 1;
	tokenerr = 0;
}

/**
 *	Parse the input file with this function
 *	@return tok The current token read
 */
char* gettok()
{
	static char line[1000];
	char *t = NULL;
	
	if (!firsttoken)
		// space, \f, \n, \r, \t, and \v are delimeters for strtok
		t = strtok(NULL, " \f\n\r\t\v");// t contains the pointer to the last token in the string
	
	while (t == NULL ? 1 : t[0] == '#')	// what is the t[0] for?
	{
		firsttoken = 0;			// toggle the flag for the first token
		// fgets reads characters from tokenfile and stores them in line until 999 characters
		// have been read or a newline or EOF character is reached
		if(fgets(line, 1000, tokenfile) == NULL)	// nothing in tokenfile
		{
			t = "";
			tokenerr = 1;
		}
		else
			t = strtok(line, " \t\n");	// t contains the pointer to the first token in the string
	}
	strcpy(tok, t);	// copy the token into tok
	return(tok);	// return the token just read
}

/**
 * Seems to return false if we have an invalid level and otherwise true
 */
int istowerinlevel(int level, int tower)
{
	int cyc,wr,thistow;
	
	if (level<0 || level>=numlevels)	// invalid level
		return 0;
	
	for (cyc=0;cyc<numcycles[level];cyc++)	// go through all of the cycles in each level
	{
		// go through each one of the numbers in the cycle
		for (wr=0,thistow=cyclestarts[level][cyc];wr<cyclesizes[level][cyc];wr++,thistow=verts[level][thistow].right) 
		{
			if (thistow==tower)	// if one of the numbers in the cycle is the same as tower
				return 1;
		}
	}
	return 0;	// This return should never be reached, but it's here to get rid of any warnings
}

/**
 *	\todo What is this for?
 */
int checkadj(int level,int tow1,int tow2)
{
	if (tow1==-1 || tow2==-1)	// we don't have a cycle (I think)
		return 0;
	if (!istowerinlevel(level,tow1))// level is an invalid level
		return 0;
	
	if (tow2==verts[level][tow1].left)	// tow2 is the left neighbor of tow1
		return -1;	
	if (tow2==verts[level][tow1].right)	// tow2 is the right neighbor of tow1
		return 1;
	
	return 0;
}

void figgerorients();	// definition
void storepatches();	// definition
void makepatchlists();	// definition

void writefile(char *readinfile, char *writeoutfile)
{

	FILE *rp, *wp;
	int lev,whichcycle,whichref,coord,tnum, whichpt, whichtow;
	if(writeoutfile==NULL)		// the user didn't enter a file to write to
		return;		
	rp = ropen(readinfile, 0);	// open the file for reading only
	tokensetup(rp);
	wp = ropen(writeoutfile, 1);	// open the file for writing only
	fprintf(wp, "\n");
	for (lev=0,gettok();!tokenerr;lev++)
	{

		for (whichcycle=0;tok[0]=='(' || tok[0]=='[';whichcycle++) // go through the cycles in the sps files
		{
			fprintf(wp, "( ");
			for (whichref=0;gettok(),tok[0]!=')' && tok[0]!=']';whichref++) // get each # in the cycle
			{
				tnum=atoi(tok);	// convert the token into an integer
				fprintf(wp, " %i", tnum);
			}
			fprintf(wp, " ) ");
			gettok();
		}
		fprintf(wp, "\n");
		for(whichpt=0;!tokenerr && tok[0]!='(' && tok[0]!='[';whichpt++)
		{
			whichtow = atoi(tok);
			fprintf(wp, "%i ", whichtow);
			gettok();
			for (coord=0;coord<3;coord++)
			{
				if(coord==2)
					fprintf(wp, " %0.2f ", points[lev][whichpt][coord]);
				else
					fprintf(wp, " %0.4f ", points[lev][whichpt][coord]);
				gettok();
			}
			fprintf(wp, "\n");
		}
	}
	rclose(rp);
	rclose(wp);
}


/*
 *	@param filename The name of the file that we need to read in
 */
void readfile(char* filename)
{
	FILE *fp;
	int lev,whichcycle,whichref,whichtow,coord,tnum,iscycle;
	int firsttow,lasttow;
	int whichpt;
	
	fp = ropen(filename, 0);// fp will hold the pointer to the input we want and it will be read-only
	tokensetup(fp);		// set a flag that we haven't looked at a token yet, etc.
	
	maxtowernum=0;
	
	for (lev=0,gettok();!tokenerr;lev++)
	{
		for (whichcycle=0;tok[0]=='(' || tok[0]=='[';whichcycle++) // go through the cycles in the sps files
		{
			iscycle=(tok[0]=='(');
			for (whichref=0;gettok(),tok[0]!=')' && tok[0]!=']';whichref++) // get each # in the cycle
			{
				tnum=atoi(tok);		// convert the token into an integer
				if (tnum>maxtowernum)	// what is maxtowernum?
					maxtowernum=tnum;
				if (whichref==0)
					firsttow=tnum;	// firsttow is the first number in the cycle
				else
				{
					// the current token's left neighbor is lasttow
					verts[lev][tnum].left=lasttow;
					// the last token's right neighbor is the token
					verts[lev][lasttow].right=tnum;	
				}
				lasttow=tnum;	// used for the next tokens left neighbor
			}
			cyclesizes[lev][whichcycle]=whichref;	// store the cycle size
			cyclestarts[lev][whichcycle]=firsttow;	// store the number that the cycle began with
			if (iscycle) 
			{					
				// the first number's left neighbor is the last number in the cycle
				verts[lev][firsttow].left=lasttow;
				// the last number's right neighbor is the first number in the cycle
				verts[lev][lasttow].right=firsttow;	
			}
			else 
			{
				verts[lev][firsttow].left=-1;	// the first number does not have a left neighbor									
				verts[lev][lasttow].right=-1;	// the last number does not have a right neighbor
			}
			gettok();
		}
		numcycles[lev]=whichcycle;	// store the number of cycles in this level
		for (whichpt=0;!tokenerr && tok[0]!='(' && tok[0]!='[';whichpt++) 
		{
			whichtow=atoi(tok);
			verts[lev][whichtow].point=whichpt;	// will be used as an index into points
			gettok(); 
			for (coord=0;coord<3;coord++) 
			{
				points[lev][whichpt][coord]=atof(tok);	// store the location of the point
				gettok();
			}
			points[lev][whichpt][3]=0.;	// time value equals 0?
			pls[lev][whichpt]=NULL;		// just initialization?
		}

		ptsperlevel[lev]=whichpt;	// store the number of points on this level
	}
	numlevels=lev;		// store the total number of levels
	
	rclose(fp);		// close the file stream
	
	figgerorients();
	storepatches();
	makepatchlists();
}

/**
 *	Calculate orientation changes on vertical edges.
 */
void figgerorients()
{
	int lev,cyc,ref,tow;
	
	for (lev=0;lev<numlevels;lev++)		// go through each level
	{
		for (cyc=0;cyc<numcycles[lev];cyc++)	// go through each cyle 
		{
			// go through each number in the cycle
			for (ref=0,tow=cyclestarts[lev][cyc];ref<cyclesizes[lev][cyc];
					ref++,tow=verts[lev][tow].right) 
			{
				// if this is the first level in the sps file, then set the up orientation to 0
				if (!istowerinlevel(lev-1,tow))		
					verts[lev][tow].up_or=0;
				else
					verts[lev][tow].up_or=verts[lev-1][tow].down_or;
				// if this is the last level in the sps file, then set the down orientation to 0
				if (!istowerinlevel(lev+1,tow))		
					verts[lev][tow].down_or=0;
				// if the left or right neighbor on this level is the 
				// same as one level up, then set the down orientation to 1
				else if (verts[lev][tow].left==verts[lev+1][tow].left || 
						verts[lev][tow].right==verts[lev+1][tow].right) 
					verts[lev][tow].down_or=1; /* True even if adj tower == -1 */
				// if the left neighbor of this level is the same as 
				// the right neighbor on the next level up or the right neighbor of 
				// this level is the same as the left neighbor of one level up, 
				// then set the down orientation equal to -1
				else if (verts[lev][tow].left==verts[lev+1][tow].right || 
						verts[lev][tow].right==verts[lev+1][tow].left) 
					verts[lev][tow].down_or=-1;
				// none of the neighbors match up and we're not on the last level of the sps file
				else	
				{
					/* Here we're on top of 2 saddles -- shouldn't happen a whole lot. !! */
					fprintf(stderr, "%s: Help! I found 2 adjacent saddles! Aaaarrgh!!\n","myname?"); /* !! */
					verts[lev][tow].down_or=1; /* So ... we fake it */
				}
			}
		}
	}
}

/**
 *	\todo What is this for?
 */
int movearound(int oldlev,int oldtower,int or,int newlev,int dir)
{
	int newtower;
	
	if (!istowerinlevel(newlev,oldtower)) 
		return -1;
	
	while(oldlev!=newlev && or!=0) 
	{
		if (oldlev<newlev) 
		{
			or *= verts[oldlev][oldtower].down_or;
			oldlev++;
		}
		else 
		{
			or *= verts[oldlev][oldtower].up_or;
			oldlev--;
		}
	}
	if (or==0) 
		return -1;
	
	or*=dir;
	if (or==1)
		newtower=verts[newlev][oldtower].right;
	else
		newtower=verts[newlev][oldtower].left;

	return newtower;
}

/**
 *	Note: Does nothing
 */
void storecap(int level,int cycle,int direction)
{
/*
!!
numpatches++;
	*/
}


void storesquare(int level,int tow1,int tow2)
{
	int plev,leftright,newlev,newtow;
	
	leftright=checkadj(level,tow1,tow2);
	patches[numpatches]=(struct patchrec *)allocate(sizeof(struct patchrec));
	patches[numpatches]->patchtype=0;
	patches[numpatches]->patchflags=
		LEFTBDRY*(movearound(level,tow1,leftright,level,-1)==-1)
		+RIGHTBDRY*(movearound(level,tow2,leftright,level,1)==-1)
		+TOPBDRY*(!istowerinlevel(level-1,tow1))
		+BOTTOMBDRY*(!istowerinlevel(level+2,tow1));
	patches[numpatches]->numpatchpts=16;
	patches[numpatches]->levs=(int*)allocate(sizeof(int)*16);
	patches[numpatches]->tows=(int*)allocate(sizeof(int)*16);
	patches[numpatches]->pd=NULL;
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=movearound(level,tow1,leftright,newlev,-1);
		patches[numpatches]->levs[plev+0*4]=newlev;
		patches[numpatches]->tows[plev+0*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=tow1;
		patches[numpatches]->levs[plev+1*4]=newlev;
		patches[numpatches]->tows[plev+1*4]=newtow;
	}
	for (plev=0;plev<4;plev++)
	{
		newlev=plev+level-1;
		newtow=tow2;
		patches[numpatches]->levs[plev+2*4]=newlev;
		patches[numpatches]->tows[plev+2*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=movearound(level,tow2,leftright,newlev,1);
		patches[numpatches]->levs[plev+3*4]=newlev;
		patches[numpatches]->tows[plev+3*4]=newtow;
	}
	
	numpatches++;
}


void storesaddle(int level,int tow1,int tow2,int tow3,int tow4)
{
	int plev,lr12,lr34,newlev,newtow;
	
	lr12=checkadj(level,tow1,tow2);
	lr34=checkadj(level,tow3,tow4);
	patches[numpatches]=(struct patchrec *)allocate(sizeof(struct patchrec));
	patches[numpatches]->patchtype=2;
	/*
	patches[numpatches]->patchflags=
	LEFTBDRY*(movearound(level,tow1,leftright,level,-1)==-1)
	+RIGHTBDRY*(movearound(level,tow2,leftright,level,1)==-1)
	+TOPBDRY*(!istowerinlevel(level-1,tow1))
	+BOTTOMBDRY*(!istowerinlevel(level+2,tow1));
	*/
	/* We assume the saddle is not on the boundary !! */
	patches[numpatches]->numpatchpts=32;
	patches[numpatches]->levs=(int*)allocate(sizeof(int)*32);
	patches[numpatches]->tows=(int*)allocate(sizeof(int)*32);
	patches[numpatches]->pd=NULL;
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=movearound(level,tow1,lr12,newlev,-1);
		patches[numpatches]->levs[plev+0*4]=newlev;
		patches[numpatches]->tows[plev+0*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=tow1;
		patches[numpatches]->levs[plev+1*4]=newlev;
		patches[numpatches]->tows[plev+1*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=movearound(level,tow2,lr12,newlev,1);
		patches[numpatches]->levs[plev+2*4]=newlev;
		patches[numpatches]->tows[plev+2*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=tow2;
		patches[numpatches]->levs[plev+3*4]=newlev;
		patches[numpatches]->tows[plev+3*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=movearound(level,tow3,lr34,newlev,-1);
		patches[numpatches]->levs[plev+4*4]=newlev;
		patches[numpatches]->tows[plev+4*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=tow3;
		patches[numpatches]->levs[plev+5*4]=newlev;
		patches[numpatches]->tows[plev+5*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=movearound(level,tow4,lr34,newlev,1);
		patches[numpatches]->levs[plev+6*4]=newlev;
		patches[numpatches]->tows[plev+6*4]=newtow;
	}
	for (plev=0;plev<4;plev++) 
	{
		newlev=plev+level-1;
		newtow=tow4;
		patches[numpatches]->levs[plev+7*4]=newlev;
		patches[numpatches]->tows[plev+7*4]=newtow;
	}
	
	numpatches++;
}

void storepatches()
{
	int lev,cyc,ref;
	int thistow,nexttow;
	int t1,t2,adjtype;
	
	numpatches=0;
	
	for (lev=0;lev<numlevels;lev++)		// cycle through all of the levels
	{
		for (cyc=0;cyc<numcycles[lev];cyc++)	// go through all of the cycles on each level
		{
			if (!istowerinlevel(lev-1,cyclestarts[lev][cyc])) 
				storecap(lev,cyc,-1);
			if (!istowerinlevel(lev+1,cyclestarts[lev][cyc]))
				storecap(lev,cyc,1);
			else 
				// go through each number in the cycle
				for (ref=0,thistow=cyclestarts[lev][cyc];ref<cyclesizes[lev][cyc];
						ref++,thistow=verts[lev][thistow].right) 
				{
					if((nexttow=verts[lev][thistow].right)==-1) 
						continue;
					if (checkadj(lev+1,thistow,nexttow)) 
						storesquare(lev,thistow,nexttow);
					else 
					{
						adjtype = checkadj(lev,thistow,nexttow);
						t1=movearound(lev,thistow,adjtype,lev+1,1);
						t2=movearound(lev,nexttow,adjtype,lev+1,-1);
						if (adjtype=checkadj(lev,t1,t2)) 
						{
							if ((adjtype==1?t1:t2)>thistow) 
								storesaddle(lev,thistow,nexttow,t2,t1);
						}
					}
				}
		}
	}
}


void makepatchlists()
{
	int pat,pt,lev,tow;
	struct patchlist **head;
	struct patchlist *oldlist;
	
	for (pat=0;pat<numpatches;pat++) 
	{
		for (pt=0;pt<patches[pat]->numpatchpts;pt++) 
		{
			lev=patches[pat]->levs[pt];
			tow=patches[pat]->tows[pt];
			if (istowerinlevel(lev,tow)) 
			{
				head=&(pls[lev][verts[lev][tow].point]);
				oldlist=(*head);
				(*head)=(struct patchlist *)allocate(sizeof(struct patchlist));
				(*head)->patchnum=pat;
				(*head)->next=oldlist;
			}
		}
	}
}


void setuppatch(int whichpatch)
{
	int pt,lev,tow,coord;
	float squarepts[16][4];
	float saddlepts[32][4];
	
	switch(patches[whichpatch]->patchtype) 
	{
		case 0: /* Square */
			for (pt=0;pt<16;pt++) 
			{
				lev=patches[whichpatch]->levs[pt];
				tow=patches[whichpatch]->tows[pt];
				if (istowerinlevel(lev,tow)) 
				{
					for (coord=0;coord<4;coord++) 
						squarepts[pt][coord]=points[lev][verts[lev][tow].point][coord];
				}
				else
					for (coord=0;coord<4;coord++) 
					{
						squarepts[pt][coord]=0.;
					}
			}
			calcsquare(patches[whichpatch]->patchflags,squarepts, &(patches[whichpatch]->pd));
			break;
		case 1: /* Cap */
				/* !! */
			break;
		case 2: /* Saddle */
			for (pt=0;pt<32;pt++) 
			{
				lev=patches[whichpatch]->levs[pt];
				tow=patches[whichpatch]->tows[pt];
				if (istowerinlevel(lev,tow)) 
					for (coord=0;coord<4;coord++) 
						saddlepts[pt][coord]=points[lev][verts[lev][tow].point][coord];
				else
					for (coord=0;coord<4;coord++)
						saddlepts[pt][coord]=0.;
			}
			calcsaddle(saddlepts,&patches[whichpatch]->pd);
	}
}


void setup()
{
	int pat;
	
	for (pat=0;pat<numpatches;pat++)
		setuppatch(pat);
}


void setupone(int level, int pt)
{
	struct patchlist *ptr;
	
	for (ptr=pls[level][pt];ptr!=NULL;ptr=ptr->next)
		setuppatch(ptr->patchnum);
}


void drawobject()
{
	int pat;
	
	for (pat=0;pat<numpatches;pat++) 
	{
		switch(patches[pat]->patchtype) 
		{
			case 0: /* square */
				drawsquare(patches[pat]->pd);
				break;
			case 1: /* cap */
				//drawcap(patches[pat]->pd);
				break;
			case 2: /* saddle */
				drawsaddle(patches[pat]->pd);
				break;
		}
	}
}

#endif	// TOPO_H
